
#include "first.h"

#include "std_opt.h"
#include "raster_util.h"
#include "clut.h"
#include <assert.h>
#include "rast_i.h"
#include "sbuf_i.h"


using namespace raster;


extern "C" void __fastcall sb_fill_afs(rend_data *rd, u32 * end_span);
extern "C" void __fastcall sb_fill_alt(rend_data *rd, u32 * end_span);
extern "C" void __fastcall sb_fill_als(rend_data *rd, u32 * end_span);
extern "C" void __fastcall sb_fill_alt_tp(rend_data *rd, u32 * end_span);
extern     void __fastcall sb_fill_pls(rend_data * _rd,u32 * end_span);



static int   cur_y;
static float fcur_y;



poly p_bot;
poly p_top;

static rend_data p_bot_rd, p_top_rd;

//extern std_fill_func ** cur_sb_plt_draw_func;


//
// todo: convert from insert sort
//
static void push_tp_poly(sbt_poly * p)
{
 float sort_z = p->sort_z;
 sbt_poly ** insp = &tp_polys;
 for(;;)
 {
  sbt_poly * cur = *insp;
  if( !cur)
   break;
  if( (int&)cur->sort_z >= (int&)sort_z )
    break;
  insp = &cur->next;
 }
 p->next = *insp;
 *insp = p;
}





// terminals
sb_arc  a_left, a_right;

static sbt_poly * active_transparent_polys;



#define TERM_L_X  (-2147483647L - 1)
#define TERM_R_X  (2147483647L)
// clippers


//
// esi is current arc
// ecx is last x
//
void  __declspec(naked)  __fastcall  step_and_sort_arcs()
{
  __asm
   {
    push esi;
    push ebp;

    push ebx;
    push edi;
    
    // initialize values
    mov esi, [a_left] sb_arc.next
    mov ecx, TERM_L_X;


    jmp TEST_NEXT;

    NEXT_ARC:

    mov eax, [esi ] sb_arc.x
    mov edx, [esi ] sb_arc.dx

    add eax, edx;   update x
    mov edi, esi;   save current  arc
 
    cmp eax,ecx
    mov esi,[esi ] sb_arc.next

    mov [edi] sb_arc.x, eax; store new x
    jl  NEED_SEARCH;
 
    mov ecx,eax;

TEST_NEXT:
    cmp esi, offset a_right;
    
    jne NEXT_ARC;


    ; done
    pop edi
    pop ebx
    pop ebp
    pop esi
    
    ret


    NEED_SEARCH:

    mov ebp, [edi ] sb_arc.prev ; get previous
    ; nop

    ; remove self
    mov [esi ] sb_arc.prev, ebp;
    ; stall

    
    mov [ebp ] sb_arc.next, esi;
    mov ebp, [ebp ] sb_arc.prev
    
    ; agi

    TRY_NEXT:
    mov edx, [ebp ] sb_arc.x
    mov ebx, [ebp ] sb_arc.prev

    cmp eax,edx;
    jge DONE0;

    mov edx, [ebx ] sb_arc.x
    mov ebp, [ebx ] sb_arc.prev
    
    cmp eax,edx;
    jl  TRY_NEXT;

    // insert self
    mov edx, [ebx ] sb_arc.next
    mov [ebx ] sb_arc.next, edi;

    mov [edi ] sb_arc.prev, ebx; 
    cmp esi, offset a_right;

    mov [edx ] sb_arc.prev, edi;
    mov [edi ] sb_arc.next, edx; 
  
    jne NEXT_ARC;

    ; done
    pop edi
    pop ebx
    pop ebp
    pop esi
    
    ret


    DONE0:

    // insert self
    mov edx, [ebp ] sb_arc.next
    mov [ebp ] sb_arc.next, edi;

    mov [edi ] sb_arc.prev, ebp; 
    cmp esi, offset a_right;

    mov [edx ] sb_arc.prev, edi;
    mov [edi ] sb_arc.next, edx; 

    jne NEXT_ARC;

    ; done
    pop edi
    pop ebx
    pop ebp
    pop esi
    
    ret
   }
  }




sbt_poly *      transparent_poly_starts[max_y_res+1];





#define FLT_MAX 	3.402823466e+38F	/* max value */



// called once at startup
void sweep_init()
{
  p_bot.reset(&p_bot_rd);
  p_top.reset(&p_top_rd);

  assert( p_bot_rd.span_buf == &span_block_stk[0]);
  assert( p_top_rd.span_buf == &span_block_stk[1]);

  p_top_rd.z0 =  FLT_MAX;
  p_bot_rd.z0 = -FLT_MAX;


  p_top_rd.dz[0] = 0.f;
  p_top_rd.dz[1] = 0.f;
  p_bot_rd.dz[0] = 0.f;
  p_bot_rd.dz[1] = 0.f;
  
  p_bot.prev =0;
  p_top.next =0;
  
  a_left.x  = TERM_L_X; 
  a_right.x = TERM_R_X; 

  

  a_left.dx = 0; 
  a_right.dx = 0; 

  a_left.prev =0;
  a_right.next =0;


  poly_transes.base[0].set(&p_bot, -1);

}









void draw_polys_lists()
{

  {
  #define DRAW_PLIST(LST,FUNC)                       \
   {                                                 \
    poly_ll_link *  tmp = LST;                       \
    LST =0;                                          \
    for(;tmp;tmp = tmp->next)                        \
     {                                               \
       poly * p = tmp->p;                         \
       rend_data * rd = p->dat;                      \
                                                     \
       if( p->cur_span != &rd->span_buf->spans[0] )  \
        {                                            \
          FUNC;                      \
        }                                            \
      }                                              \
   }


   DRAW_PLIST(als_polys,sb_fill_als(rd,p->cur_span))
   DRAW_PLIST(alt_polys,sb_fill_alt(rd,p->cur_span))
   DRAW_PLIST(afs_polys,sb_fill_afs(rd,p->cur_span))
   DRAW_PLIST(plt_polys, ((rend_data_tex_fn*)rd)->fill_func(rd,p->cur_span))
 //  DRAW_PLIST(pls_polys,sb_fill_pls(rd,p->cur_span))

   
 
  }

  


}



/*
void draw_polys_and_reset_bufs()
 {
 TM_START( raster );

#ifdef PROF
  p_buf_used =  (int)sb_stk_ptr - (int)span_block_stk;
  e_buf_used =  (int)sbuf_stk.p - (int)sbuf_stk.base;
#endif


   // fill in black
#ifdef CLEAR_SCREEN
	fill_black(&p_bot.span_buf.spans[0],p_bot.cur_span );
#endif


if(display_miser::bytes_per_pixel == 2)
{
#define DRAW_PLIST(LST,FUNC)                       \
 {                                                 \
  poly_ll_link *  tmp = LST;                       \
  LST =0;                                          \
  for(;tmp;tmp = tmp->next)                        \
   {                                               \
     poly * p = tmp->p;                         \
     rend_data * rd = p->dat;                      \
                                                   \
     if( p->cur_span != &rd->span_buf->spans[0] )  \
      {                                            \
        FUNC(rd,p->cur_span);                      \
      }                                            \
    }                                              \
 }


 DRAW_PLIST(als_polys,draw_sp_gour)
 DRAW_PLIST(alt_polys,draw_sp_lin_tex)
 DRAW_PLIST(afs_polys,fill_sp_f0)
 DRAW_PLIST(g3_polys,fill_g3)

// DRAW_PLIST(plt_polys,draw_persp_tex)

}
else
{
// Dex("todo");
}

  
   poly_ll_links.make_empty();
   
   poly_edge_stk.make_empty();
   rend_dat_stk.make_empty();
   empty_block_stk();
  reset_y_extents();


  TM_END(raster);

 }


  */



/*void do_trans(int x)
 {
   sb_arc * tmp = cur_arc;

   assert( search_polys.is_empty() );

#ifdef PROF
   {
	 ++num_trans;
     if( FX1.d(tmp->next->x) == x )
	    {
       ++num_atrans;		  
	    }

   }
#endif

    do
    {

         // is it an interior arc?
         if( tmp->type == 2 )
          {
           poly * lp = ((sb_arc_i*)tmp)->lp;
           poly * rp = ((sb_arc_i*)tmp)->rp;
     
         
           // case 0 : (most common case)
           // left active, right inactive
           if(  POLY_ACTIVE(lp->next)&& ( POLY_READY(rp->next) ) )
            {
 #ifdef PROF
            ++good_interiors;
 #endif
             // lp->next might not be best cmp.
             // swap
             dlist::link(rp,lp->next);
             dlist::link(lp->prev,rp);
             lp->next=0;
            }
           else 
           {
#ifdef PROF
          ++bad_interiors;
#endif
            // pretend that we are 2 arcs
             if( POLY_ACTIVE(lp->next) )
             {
              dlist::remove(lp);
              lp->next=0;
             }
             else
             {
              INC_NEG_COUNT(lp->next);
             }
           
            search_polys.push().set(rp);
           }

          } // if type == 2
         else if( tmp->type == 1)
          {
           poly * sp = ((sb_arc_r*)tmp)->lp;
     
           // is it active?
           if( POLY_ACTIVE( sp->next ) )
            {
             dlist::remove(sp);
             sp->next=0;
            }
           else
            {
             INC_NEG_COUNT(sp->next); // increment hidden neg span counter
            }
         }  // type == 1
          else // type == 1
          {
            assert( tmp->type == 0);
            search_polys.push().set( ((sb_arc_l*)tmp)->rp,((sb_arc_l*)tmp) );
           }
       
         tmp = tmp->next;
       
	   } while( FX1.d(tmp->x) == x );
      

       cur_arc = tmp;



    while(!search_polys.is_empty() )
     {
    
       insert_ply & ip = search_polys.pop();
       poly * sp = ip.ply;
       
	   // finished before we start?
       if( !POLY_READY(sp->next) )
       {
         DEC_NEG_COUNT( sp->next);
         continue;
       }

        if( ip.larc && ip.larc->l_prev && ( ip.larc->l_prev->next == ip.larc->l_next  ) )
         { 
           poly * l_prev = ip.larc->l_prev;
           poly * l_next = ip.larc->l_next;
             // same place?
            assert( l_next->prev == l_prev );
      
            dlist::link(sp,l_next);
            dlist::link(l_prev,sp);
          #ifdef PROF
            ++cont_count;
          #endif
         continue;
       }

       #ifdef PROF
             ++search_count;
       #endif

       poly * ins = p_top.prev;
             
       float sz = sp->z_at(x,cur_y);
        // find position in list
        for(;;)
        {
          float tz = ins->z_at(x,cur_y);
          if( tz >= sz )
            break;
           ins = ins->prev;
        }
         dlist::link(sp,ins->next);
         dlist::link(ins,sp);

         // cache position
         if(ip.larc)
          {
           ip.larc->l_prev = sp->prev;
           ip.larc->l_next = sp->next;
          }
 
      }



 }

 
  */















   


struct insert_ply
 {
  poly *ply;
  sb_arc_l * larc; // might be 0
  inline void set(poly*_ply,sb_arc_l * _larc=0)
   {
    ply = _ply;
    larc = _larc;
   }
 };
const int max_search_polys = 512;

static min_stack<insert_ply,max_search_polys> search_polys;

//
// visible poly 
//
struct visible_poly_span
{
  raster::poly * poly;
  int x0, x1; // x1 is always  > x0
};
const int max_visible_poly_spans = 1024;
static min_stack<visible_poly_span,max_visible_poly_spans> visible_poly_spans;

                                      
       // remove all active polys that end here
#define OVERLAP_CNT 1024

#define POLY_ACTIVE(PTR)   ( (u32)(PTR) > OVERLAP_CNT )
#define POLY_READY(PTR)    ( (PTR) ==0 )
#define INC_NEG_COUNT(PTR) (  ++ ((u32 &) (PTR) ) )
#define DEC_NEG_COUNT(PTR) (  -- ((u32 &) (PTR) ) )


//static sb_arc * cur_arc;
static int cur_x;

void __fastcall do_poly_insertion(poly * sp,sb_arc_l * larc )
{
__debug(
   assert( p_top.next == 0 );
   assert_valid(sp);
   {
     poly * ins = p_top.prev;
     for(;ins != &p_bot;ins = ins->prev)
     {
      assert_valid(ins);
     }

   })

   assert( search_polys.p < &search_polys.base[max_search_polys] );    

	   // finished before we start?
       if( !POLY_READY(sp->next) )
       {
         if( !POLY_ACTIVE(sp->next)  )
         {
          DEC_NEG_COUNT( sp->next);
         }
         return;
       }


       #ifdef PROF
             ++search_count;
       #endif

       poly * ins = p_top.prev;
       if( ins != &p_bot )
       {
        assert_valid(ins);

        float sz = sp->z_at(float(cur_x),fcur_y);
        // find position in list
        for(;;)
        {
          if( ins == &p_bot )
            break;
          float tz = ins->z_at(float(cur_x),fcur_y);
          if( tz <= sz )
            break;
           ins = ins->prev;
        }
       }

   assert( p_top.next == 0 );

       dlist::link(sp,ins->next);
       dlist::link(ins,sp);

         // cache position
        if(larc)
         {
           larc->l_prev = sp->prev;
           larc->l_next = sp->next;
          }


   assert( p_top.next == 0 );
 
}

       
min_stack<poly_trans,max_poly_transes> poly_transes;

// data from the last poly/
static poly * last_top;


void __declspec(naked) __fastcall  do_all_trans(sb_arc *)
 {
  __asm
  {
    push ebx
    push ebp
   
    mov ebx, [ecx] sb_arc.x
    push esi
 
    sar ebx, 16;
    push edi

    ; save last top
    mov [last_top], offset p_bot;
    mov edx, [clip_x_max];

    jmp CHECK_LINE_FINISHED;

    
    ; 
    ; ebx has cur x
    ;
NEXT_ARC:

    ; remove!!
   ;  mov [cur_x], ebx;



    mov eax, [ecx] sb_arc.type; 
    mov edx, [ecx] sb_arc_i.lp; 

    dec eax;
    jle  NOT_MIDDLE;
     
    mov eax, [ecx] sb_arc_i.rp;
    mov ebx, [edx] poly.next; 

    ;edx is lp
    ;eax is rp
    ;ebx is lp->next

    cmp ebx, OVERLAP_CNT;
    jb  BAD_INTERIOR_LEFT_INACT;
  
    ;edx is lp
    ;eax is rp
    ;ebx is lp->next
    ;edi is lp->prev
    ;esi is rp->next

    mov esi, [eax] poly.next;
    mov edi, [edx] poly.prev;

    test esi,esi;
    jne  BAD_INTERIOR_LEFT_ACT;
   
    ; swap lp with rp in poly list
    mov [ebx] poly.prev, eax;
    mov ecx, [ecx] sb_arc.next; 
   
    mov [eax] poly.next, ebx;
    mov [eax] poly.prev, edi;
    
    mov [edi] poly.next, eax;
    mov ebx, [ecx] sb_arc.x

    mov [edx] poly.next, esi; esi is 0
    
    
    
    
    jmp CUR_ARC_DONE;


 

     ; edx is lp
     ; ebx is lp->next

; left poly is inactive, add 1 to its inactiveness  
BAD_INTERIOR_LEFT_INACT:


     inc  ebx;
     mov  esi, [search_polys.p];

     mov  [edx] poly.next, ebx;
     jmp   ADD_RP_TO_SEARCH;

     ; edx is lp
     ; ebx is lp->next
     ; edi is lp->prev

BAD_INTERIOR_LEFT_ACT:

     ; need to remove lp
     mov esi, [search_polys.p];
     mov [ebx] poly.prev, edi

     mov [edi] poly.next, ebx
     mov dword ptr [edx] poly.next, 0;

 ADD_RP_TO_SEARCH:

     mov dword ptr [esi] insert_ply.larc, 0;

     mov ecx,[ecx] sb_arc.next 
     mov [esi] insert_ply.ply , eax;

     add esi, 8;
     mov ebx, [ecx] sb_arc.x 

     mov [search_polys.p], esi;
     jmp CUR_ARC_DONE


  NOT_MIDDLE:

  mov eax, [edx] poly.next;
  jl LEFT_ARC;
  ; we are a right arc


  ; ecx cur arc
  ; edx lp
  ; eax next
  ; edi prev

  mov edi, [edx] poly.prev;
  mov ecx,[ecx] sb_arc.next 

  cmp  eax,OVERLAP_CNT
  jb  RP_NOT_ACTIVE;


  mov ebx, [ecx] sb_arc.x
  mov [eax] poly.prev, edi;

  mov dword ptr [edx] poly.next, 0;
  mov [edi] poly.next, eax;

  
 

  jmp CUR_ARC_DONE



 RP_NOT_ACTIVE:
  
  inc eax;
  mov ebx, [ecx] sb_arc.x

  mov  [edx] poly.next, eax;


  jmp CUR_ARC_DONE


  LEFT_ARC:
  ; edx : ply
  ; eax : ply->next
  ; ebx : l_prev
  ; edi : l_next

    mov ebx,[ecx] sb_arc_l.l_prev;

    test eax,eax;
    jnz  POLY_NOT_READY;

    mov esi,[ebx] poly.next
    mov edi,[ecx] sb_arc_l.l_next;
    
    cmp edi, esi;
    jne  NOT_SAME_PAIR;

    mov ecx,[ecx] sb_arc.next 

    mov [ebx] poly.next, edx;
    mov [edx] poly.prev, ebx;

    mov [edi] poly.prev, edx;
    mov ebx, [ecx] sb_arc.x

    mov [edx] poly.next, edi;
    jmp CUR_ARC_DONE;



POLY_NOT_READY:
 
  
   dec eax;
   mov ecx,[ecx] sb_arc.next 

   cmp eax, OVERLAP_CNT;
   ja  __overlapping;

   mov [edx] poly.next,eax;
   nop


 __overlapping:
   
   mov ebx, [ecx] sb_arc.x
   jmp CUR_ARC_DONE;
   



NOT_SAME_PAIR:

   mov esi, [search_polys.p];
   mov edi, ecx
 
   mov ecx,[ecx] sb_arc.next 
   nop
    
   mov [esi] insert_ply.ply , edx;
   mov [esi] insert_ply.larc, edi;

   add esi, 8;
   mov ebx, [ecx] sb_arc.x

   mov [search_polys.p], esi;
  ;   jmp CUR_ARC_DONE;


CUR_ARC_DONE:
  
 
   sar ebx,16;
   mov  esi, [search_polys.p];
    
   cmp ebx,ebp;
   je  NEXT_ARC_SK;




   cmp esi, offset search_polys.base;
   jne  NEED_INSERTS;

  INSERTS_DONE:
   

   mov  eax, [last_top];
   mov  edi, [ p_top ] poly.prev;

   mov esi, [poly_transes.p];
   ; is there a new top?
   cmp eax, edi;

   mov edx, [clip_x_max];
   je  CHECK_LINE_FINISHED;


   mov [last_top], edi;
   mov [esi] poly_trans.x, ebp;
   mov [esi] poly_trans.new_top, edi;
   add esi, 8;
   
   mov [poly_transes.p], esi;
   
  CHECK_LINE_FINISHED:

   cmp ebx, edx;
   jge FINISHED_LINE;

   mov ebp, ebx;

NEXT_ARC_SK:
   jmp NEXT_ARC;


NEED_INSERTS:

   push ecx;  save cur arc
   mov [cur_x], ebp;

NEXT_INSERT:

   mov ecx, [esi-8] insert_ply.ply;
   mov edx, [esi-8] insert_ply.larc;


   sub esi, 8;
   call do_poly_insertion;

   cmp esi, offset search_polys.base;
   jne NEXT_INSERT;
   
   ;restore arc
   pop ecx;
   mov [search_polys.p],esi
  
   jmp  INSERTS_DONE;


 FINISHED_LINE:

    pop edi
    pop esi

    pop ebp
    pop ebx
  

    ret

 }
} 






//
//  writes out visible spans of the poly based on visible_poly_spans for
// the current scan line
//
static void __fastcall  submit_transparent_span(sbt_poly * p, int sx0, int sx1)
{
	visible_poly_span * start = visible_poly_spans.base;
	visible_poly_span * end   = visible_poly_spans.p;
	for(;start < end ;++start)
	{
		// intersection?
		// ends before us?
		if( start->x1 <= sx0 )
			 continue;

		// starts after we are done?
   if( start->x0 >= sx1 )
			 break;
		int x0 = __max( sx0, start->x0 );
		int x1 = __min( sx1, start->x1 );


		// no intersection?
		if( x0 >= x1 ) 
		 continue;
	   
		        

		// visible?

  float z0 = p->z_at(float(x0),fcur_y);
		float z1 = start->poly->z_at(float(x0),fcur_y);
		
		if( z0 > z1 )
		{
		 (p->push_span(pack_span_dat(cur_y,x0, x1)));
		}
  else
  {
   int q=0;
  }


	}


}


//
// steps all polys by 1 y at cur_y and outputs spans
//

static void update_transparent_polys()
{

 // first insert new polys
  sbt_poly * new_start = transparent_poly_starts[cur_y];
  if( new_start )
  {
	  transparent_poly_starts[cur_y]= 0;
   
   while(new_start)
	  {
		  sbt_poly * cur = new_start;
		  new_start = new_start->next;
		  llist::add_head(  active_transparent_polys, cur );
	   assert( cur->dat->z0 != 0.0f );

   }
  }


  // step all polys, removing those that are finished 
  sbt_poly ** p_cur = &active_transparent_polys;
  sbt_poly * cur = active_transparent_polys;
  for(;;)
  {

	  int x0, x1;
	  if( !cur )
		  break; // finished
      // check left side 
	  sbt_arc * left = cur->get_left_arc();

   if( left->end_y <= cur_y )
	   {
	     ++left;
     
   /*  if( cur->end_y < cur_y )
	     {
       *p_cur = cur->next;
  	    sbt_poly * tmp = cur;       
       cur = cur->next;
       push_tp_poly(tmp); // register to draw
       continue;
      } */  
	     assert( left->end_y > cur_y );
	     cur->get_left_arc() = left;
	   }
	   x0 = FX1.d(left->x);
	   left->x += left->dx;
    sbt_arc * right =  cur->get_right_arc();
    if( right->end_y <= cur_y )
	   {
	     ++right;
	     assert( right );
	     assert( right->end_y > cur_y );
	     cur->get_right_arc() = right;
	   }

    x1 = FX1.d(right->x);
	   right->x += right->dx;
	   // insert span
 	  if( x1 > x0 )
    {
     submit_transparent_span(cur,  x0, x1 );
  	 }

    // poly is finished?
	   if( cur->end_y > cur_y )
	   {
     p_cur = &cur->next;
     cur = cur->next;
    }
    else
    {
		     *p_cur = cur->next;
  	    sbt_poly * tmp = cur;       
       cur = cur->next;
       push_tp_poly(tmp); // register to draw
    }

//   p_cur = &cur->next;
//   cur = cur->next;

  } 


}



void do_sweep()
{

  // reset
  dlist::link( &a_left, &a_right );
 
  assert( a_right.x ==  TERM_R_X); 

 
  for(cur_y=min_y_span;cur_y< max_y_span;++cur_y)
   {
    fcur_y = (float)cur_y;


    // remove old arcs
    sb_arc * sb  = arc_trans[cur_y].enders;
    arc_trans[cur_y].enders=0;
    
    while(sb)
     {
       sb->remove();
       sb = sb->end_next;
     }


   assert( a_right.x ==  TERM_R_X); 
    
    // update x's and re-sort arcs
    step_and_sort_arcs();



   // insert new arcs
    sb = arc_trans[cur_y].starters;
    
    assert( a_right.x ==  TERM_R_X); 

    if( sb )
    {
      arc_trans[cur_y].starters = 0;
      sb_arc * tmp = sb->next;
      sb->next=0;
      while( tmp )
      {
       sb_arc *cur = tmp;
  
       tmp = tmp->next;
       sb_arc ** insp = &sb;

       int cx = cur->x;

       for(;;)
        {
         sb_arc * ins = *insp;
         if(!ins) 
           break;
         if( ins->x >= cx )
           break;
         insp = &ins->next;

        }
        cur->next = *insp;
        *insp = cur;
      }  // while tmp

     // merge sort them in
     sb_arc * ins = a_left.next;

     assert( a_right.x ==  TERM_R_X); 
     do
     {
      sb_arc *  cur = sb;
      sb = sb->next;

    
      int sx = cur->x;

      // walk forwards until we find a valid position
      for(;;)
       {
        if( ins->x >= sx )
          break;
        ins = ins->next;
       }
      dlist::link(ins->prev,cur);
      dlist::link(cur,ins);
     
     } while(sb);


    }// if new arcs




      // clear act poly list
     dlist::link( &p_bot, &p_top  );

     assert( a_right.x ==  TERM_R_X); 
     assert( p_top.next == 0 );

     // do transitions
     poly_transes.p = &poly_transes.base[1];
     do_all_trans(a_left.next);
   

     assert( a_right.x ==  TERM_R_X); 
     assert( p_top.next == 0 );
     
     // remove active polys if any, these only exist if
	 // polys go beyond the edge of the screen
     {
       poly * tmp = p_top.prev;
       while( tmp != &p_bot  )
       {
         tmp->next = 0;
         tmp = tmp->prev;
       }
     }

     // add terminal
     poly_transes.push().set( 0,max_x_res+1);

     poly_trans * pt = poly_transes.base;
   
     assert( a_right.x ==  TERM_R_X); 
    
	    visible_poly_spans.make_empty();
     u32 * span_p = span_buffer[cur_y];
     for(;*span_p != 0;++span_p)
     {
        int start = (*span_p) & 1023;
        int end   = (*span_p>>10);

         // skip to start;
         while(pt->x <= start )
         {
          ++pt;
         }

         int last_x = start;
         for(;;) 
         {
          int x = pt->x;
          
		        if( x >= end )
          {
           if( end > last_x )
		         {
			         pt[-1].new_top->push_span( pack_span_dat(cur_y,last_x, end) );
            visible_poly_span & sp = visible_poly_spans.push();
			         sp.x0 = last_x;
			         sp.x1 = end;
			         sp.poly = pt[-1].new_top;
		         }
		         break;
          }

          pt[-1].new_top->push_span( pack_span_dat(cur_y,last_x,x));
		        visible_poly_span & sp = visible_poly_spans.push();
		        sp.x0 = last_x;
		        sp.x1 = x;
		        sp.poly = pt[-1].new_top;
       
           #ifdef PROF
                 ++span_count;
                 span_dx += x - last_x;
           #endif

           if( x >= clip_x_max)
           {
            break;
           }
          last_x = x;
          ++pt;
         }
     }


    assert( a_right.x ==  TERM_R_X); 


 //
	// write values back into buffer
	//
	{
      span_p =span_buffer[cur_y];
      u32 * span  = &span_block_stk[0].spans[0];
      for(;;)
      {
       if(  span == p_bot.cur_span)
         break;
       if( ((int)span &63) == 0 ) 
        span = (u32*)span[-16];
       
       assert( span_p < &span_buffer[cur_y][MAX_SPANS] );

       *span_p++  = (*span++)&( 1023 + (1023<<10) );
      }
      // null terminate
      *span_p = 0;
      // reset list
      p_bot.cur_span = &span_block_stk[0].spans[0];
  }
	 assert( a_right.x ==  TERM_R_X); 

     //
  	 // handle transparent arcs
	 //
	 update_transparent_polys();

     
   }  // for y
   arc_trans[cur_y].enders =0;


   {
    sbt_poly * p =   active_transparent_polys;
    for(;p;p = p->next)
    {
     assert( p->end_y == cur_y);
    }
    active_transparent_polys=0;
   }
   	    // insert new arcs
   assert(  transparent_poly_starts[cur_y] == 0 );

   // clean up 
  assert( arc_trans[cur_y].enders == 0);
 // arc_trans[cur_y].enders =0;
  assert( arc_trans[cur_y].starters == 0 );
}



static void reset_buffers()
{
 #ifdef PROF
  p_buf_used =  (int)sb_stk_ptr - (int)span_block_stk;
  e_buf_used =  (int)sbuf_stk.p - (int)sbuf_stk.base;
#endif

   
   poly_edge_stk.make_empty();
   rend_dat_stk.make_empty();
   empty_block_stk();
   reset_y_extents();
}


void __fastcall raster::draw_transparent_polys()
{
   
 if( tp_polys )
 {
    sbt_poly *  tmp = tp_polys; 
    tp_polys =0;  
    for(;tmp;tmp = tmp->next)                        
     {                                               
       rend_data * rd = tmp->dat;                      
                                                     
       if( tmp->cur_span != &rd->span_buf->spans[0] )  
        {                                            
         tmp->fill_func(rd,tmp->cur_span);
        }                                            
      }                                              
 }


 transparent_poly_stk.make_empty();
 empty_block_stk_tp();
}


void __fastcall raster::sweep_and_draw()
{
 do_sweep();
 draw_polys_lists();
 reset_buffers();
}
